#include "nsIServiceManager.h"
#include "nsIGenericFactory.h"
#include "nsIImagePermission.h"

#include "nsIPref.h"

#include "nsCookie.h"
#include "xp_core.h"
#include "nsImagePermission.h"
#include "nsICookieService.h"
#include "nsString.h"
#include "nsIURL.h"

static NS_DEFINE_IID(kCookieServiceCID, NS_COOKIESERVICE_CID);

/* Implementation */
// private members related to properties listen
#define image_behaviorPref "network.image.imageBehavior"

// static data initialization
COOKIE_BehaviorEnum nsImagePermission::mImageBehavior = COOKIE_Accept;
nsVoidArray* mDocPermission = nsnull;
nsVoidArray* mImgPermission = nsnull;

// helper classes

MODULE_PRIVATE int PR_CALLBACK
imageBehaviorPrefChanged(const char * newpref, void * data) {
  PRInt32 n;
  nsresult rv;
  NS_WITH_SERVICE(nsIPref, prefs, "@mozilla.org/preferences;1", &rv);
  if (NS_FAILED(prefs->GetIntPref(image_behaviorPref, &n))) {
    nsImagePermission::SetBehaviorPref(COOKIE_Accept);
  } else {
    nsImagePermission::SetBehaviorPref((COOKIE_BehaviorEnum)n);
  }
  return 0;
}


NS_IMPL_ISUPPORTS1(nsImagePermission, nsIImagePermission)

nsImagePermission::nsImagePermission()
  : mInitted(PR_FALSE)
{
  NS_INIT_ISUPPORTS();
  /* member initializers and constructor code */
}

nsImagePermission::~nsImagePermission()
{
  /* destructor code */
}


nsresult nsImagePermission::Init() 
{
  if(mInitted) {
    NS_ASSERTION(0, "nsImagePermission service already initted!");
    return NS_ERROR_ALREADY_INITIALIZED;
  }

  RegisterPrefsCallbacks();
  mInitted = PR_TRUE;

  return NS_OK;
}


/* void SetPermission (in string aURL, in PRInt32 type, in boolean permission); */
NS_IMETHODIMP nsImagePermission::ChangePermission(const char *aURL, PRInt32 type, PRBool permission)
{
  if(!mDocPermission) {
    mDocPermission = new nsVoidArray();
    if(!mDocPermission) return NS_ERROR_FAILURE;
  }
  if(!mImgPermission) {
    mImgPermission = new nsVoidArray();
    if(!mImgPermission) return NS_ERROR_FAILURE;
  }
  switch(type) {
  case DOCUMENT_PERMISSION:
    ChangePermission(aURL, permission, mDocPermission);
    break;
  case IMAGE_PERMISSION:
    ChangePermission(aURL, permission, mImgPermission);
    break;
  default:
    NS_ASSERTION(0, "Undefined permission type");
    return NS_ERROR_FAILURE;
  };
 
  return NS_OK;
}

/* block certain hosts from loading images */
PRBool nsImagePermission::permitLoad(const char * image_url, const char *doc_url) {

    /* convert image_url to an nsIURL so we can extract host and scheme */
    nsresult rv;
    NS_WITH_SERVICE(nsIURL, uri, "@mozilla.org/network/standard-url;1", &rv);
    if (NS_FAILED(rv) || NS_FAILED(uri->SetSpec(image_url))) {
        return PR_TRUE;
    }

    /* extract scheme -- we block loading of only http and https images */
    char * scheme;
    rv = uri->GetScheme(&scheme);
    if (NS_FAILED(rv)) {
        return PR_TRUE;
    }
    if (PL_strcasecmp(scheme, "http") && PL_strcasecmp(scheme, "https")) {
        Recycle(scheme);
        return PR_TRUE;
    }
    Recycle(scheme);

    /* extract host */
    char * host;
    rv = uri->GetHost(&host);
    if (NS_FAILED(rv) || !host) {
        return PR_TRUE;
    }

    /*********************/
    if (NS_FAILED(rv) || NS_FAILED(uri->SetSpec(image_url))) {
	Recycle(host);
        return PR_TRUE;
    }

    /* extract host */
    char * firstHost;
    rv = uri->GetHost(&firstHost);
    if (NS_FAILED(rv) || !firstHost) {
	Recycle(host);
        return PR_TRUE;
    }

    if (!firstHost) {
        firstHost = PL_strdup(host);
    }

    /* check to see if we need to block image from loading */
    NS_WITH_SERVICE(nsICookieService, cookieservice, kCookieServiceCID, &rv);
    if (NS_FAILED(rv)) {
        Recycle(host);
        Recycle(firstHost);
        return PR_TRUE;
    }
    PRBool permission;
    rv = cookieservice->Image_CheckForPermission(host, firstHost, permission);
    Recycle(host);
    Recycle(firstHost);
    if (NS_FAILED(rv)) {
      return PR_TRUE;
    }
    return permission;
}


/* PRInt32 CheckImagePermission (in string aDocURL, in string aImgURL); */
NS_IMETHODIMP nsImagePermission::CheckImagePermission(const char *aDocURL, const char *aImgURL, PRInt32 *_retval)
{
  PRInt32 index;

  if (permitLoad(aImgURL, aDocURL)) {
      *_retval = LOAD_ALLOWED;
      return NS_OK;
  }

  //ok, if we are here then image load is disabled for some reason
  // lets check whenever image still might be shown
    if( LookupURL( aDocURL, mDocPermission, index ) ) {
      *_retval = LOAD_WHOLE_GROUP;
    } else if( LookupURL( aImgURL, mImgPermission, index) ) {
      *_retval = LOAD_ALLOWED;
    } else { 
      *_retval = LOAD_FORBIDDEN;
    }

  return NS_OK;
}


COOKIE_BehaviorEnum nsImagePermission::GetBehaviorPref()
{
  return mImageBehavior;
}

void nsImagePermission::SetBehaviorPref(COOKIE_BehaviorEnum x)
{
  mImageBehavior = x;
}

void nsImagePermission::RegisterPrefsCallbacks(void) 
{
  PRInt32 n;
  nsresult rv;
  NS_WITH_SERVICE(nsIPref, prefs, "@mozilla.org/preferences;1", &rv);

  // Initialize for cookie_behaviorPref
  if (NS_FAILED(prefs->GetIntPref(image_behaviorPref, &n))) {
    n = COOKIE_Accept;
  }
  SetBehaviorPref((COOKIE_BehaviorEnum)n);
  prefs->RegisterCallback(image_behaviorPref, imageBehaviorPrefChanged, NULL);

}
//----------------------------------------------------------------------
// helper function
void nsImagePermission::ChangePermission(const char* aURL, PRBool bAllow, nsVoidArray* pArray) {
  PRInt32 index = -1;

  PRBool bExist = LookupURL(aURL, pArray, index);
  if( bExist ) {
    if( !bAllow ) {
      // remove permission;
      nsString* pTmp = (nsString*)pArray->ElementAt(index);
      pArray->RemoveElementAt(index);
      delete pTmp;
    }
  } else {
    if( bAllow ) {
      nsString* pStr = new nsString();
      pStr->AssignWithConversion(aURL);
      pArray->AppendElement(pStr);
    }
  }
}

PRBool nsImagePermission::LookupURL(const char* aURL, nsVoidArray* pArray, PRInt32& index) {
  if(!pArray) {
    return PR_FALSE;
  }
  for(int i = 0; i < pArray->Count(); i++) {
    nsString* pStr = (nsString*)pArray->ElementAt(i);
    if(pStr->EqualsWithConversion(aURL)) {
      index = i;
      return PR_TRUE;
    }
  }
  return PR_FALSE;
}


